from collections import Counter
from unittest import TestCase

from mab_ranking.bandits.bandits import BetaThompsonSampling
from mab_ranking.bandits.rank_bandits import IndependentBandits


class IndependentBanditsTest(TestCase):
    def test_choose(self):
        num_ranks = 3
        rank_bandit = IndependentBandits(num_ranks, BetaThompsonSampling, num_arms=10)

        selected_arms = rank_bandit.choose()

        assert len(selected_arms) == num_ranks
        assert len(selected_arms) == len(set(selected_arms))

    def test_update(self):
        rank_bandit = IndependentBandits(3, BetaThompsonSampling, num_arms=8)

        selected_arms = [0, 1, 2]
        rank_bandit.update(selected_arms=selected_arms, rewards=[1.0, 0.0, 0.0])

        assert rank_bandit.rank_bandits[0].rewards == [2.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0]
        assert rank_bandit.rank_bandits[0].num_tries == [3.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0, 2.0]

    def test_choose_and_update(self):
        rank_bandit = IndependentBandits(2, BetaThompsonSampling, num_arms=4)

        chosen_arm_in_first_position = []
        for _ in range(500):
            selected_arms = rank_bandit.choose()
            first_chosen_arm = selected_arms[0]
            chosen_arm_in_first_position.append(first_chosen_arm)

            # Only arm with id 1 at 1st position receives a reward
            reward = 1.0 if first_chosen_arm == 1 else 0.0
            rank_bandit.update(
                selected_arms=selected_arms,
                rewards=[reward, 0.0]
            )

        assert Counter(chosen_arm_in_first_position).most_common(1)[0][0] == 1



from collections import Counter
from unittest import TestCase

from mab_ranking.bandits.bandits import BetaThompsonSampling


class BetaThompsonSamplingTest(TestCase):
    def test_choose(self):
        num_arms = 10
        bandit = BetaThompsonSampling(num_arms=num_arms)
        best_arm_id, sorted_arm_ids = bandit.choose()

        assert best_arm_id in range(num_arms)

        assert len(sorted_arm_ids) == len(set(sorted_arm_ids))

        for _arm_id in sorted_arm_ids:
            _arm_id in range(num_arms)

    def test_update(self):
        bandit = BetaThompsonSampling(num_arms=4)

        bandit.update(1, reward=1.0)

        assert bandit.rewards == [1.0, 2.0, 1.0, 1.0]
        assert bandit.num_tries == [2.0, 3.0, 2.0, 2.0]

    def test_choose_and_update(self):
        bandit = BetaThompsonSampling(num_arms=2)

        chosen_arms = []
        # arm id 0 provides much more reward than arm id 1
        for _ in range(500):
            best_arm_id, _ = bandit.choose()
            chosen_arms.append(best_arm_id)

            reward = 1.0 if best_arm_id == 0 else 0.0
            bandit.update(best_arm_id, reward=reward)

        chosen_arms_with_frequency = Counter(chosen_arms)
        assert chosen_arms_with_frequency[0] > chosen_arms_with_frequency[1]



from abc import ABCMeta
from abc import abstractmethod

import numpy as np


class Bandit(metaclass=ABCMeta):
    """
    Base abstract class to inherit from for Multi-Armed-Bandits implementations.

    Arm ids are 0-based indexed.
    """
    def __init__(self, num_arms):
        self.num_arms = num_arms

    @abstractmethod
    def choose(self, context=None):
        """
        Chooses an arm id

        :param context: [dict, default=None], context of the request. Only discrete context values
        are supported of type str. Example:
        {
            "context_a": "value_a1",
            "context_b: "value_b3"
        }

        :return: [tuple[int, list[int]], a tuple where the left item is the best chosen arm id
        and the right item has a list of arm ids (including the best one in the first position)
        ordered in a decreasing order in terms of best possible arm.
        """
        pass

    @abstractmethod
    def update(self, arm_id, reward, context=None):
        """
        Updates algorithm given the feedback
        :param arm_id: [int], arm id
        :param reward: [float], reward for arm id
        :param context: [dict, default=None], context of the request. Only discrete context values
        are supported of type str. Example:
        {
            "context_a": "value_a1",
            "context_b: "value_b3"
        }
        """
        pass


class BetaThompsonSampling(Bandit):

    def __init__(self, num_arms):
        super().__init__(num_arms)
        self.rewards = [1.0 for _ in range(self.num_arms)]
        self.num_tries = [2.0 for _ in range(self.num_arms)]

    def choose(self, context=None):
        probs = [
            np.random.beta(total_reward, n - total_reward)
            for n, total_reward in zip(self.num_tries, self.rewards)
        ]

        sorted_indices = np.argsort(probs)[::-1][-len(probs):].tolist()

        return sorted_indices[0], sorted_indices

    def update(self, arm_id, reward, context=None):
        self.num_tries[arm_id] += 1
        self.rewards[arm_id] += reward



from abc import ABCMeta, abstractmethod


class RankBandit(metaclass=ABCMeta):
    """
    Base abstract class to inherit from for Ranking Multi-Armed-Bandits implementations.

    Arm ids are 0-based indexed.
    """
    def __init__(self, num_ranks, bandit_class, **kwargs):
        self.num_ranks = num_ranks
        self.rank_bandits = [bandit_class(**kwargs) for _ in range(self.num_ranks)]

    @abstractmethod
    def choose(self, context=None):
        """
        Chooses an arm id for each position of the rank

        :param context: [dict, default=None], context of the request. Only discrete context values
        are supported of type str. Example:
        {
            "context_a": "value_a1",
            "context_b: "value_b3"
        }

        :return: [list[int]], a list of chosen arm id per position. It has non-duplicates.
        """
        pass

    def update(self, selected_arms, rewards, context=None):
        """
        Updates bandit algorithms

        :param selected_arms: [list[int]], a list of arm ids
        :param rewards: [list[float]], a reward per arm id in the list
        :param context: [dict, default=None], context of the request. Only discrete context values
        are supported of type str. Example:
        {
            "context_a": "value_a1",
            "context_b: "value_b3"
        }
        """
        for i, arm_reward_tuple in enumerate(zip(selected_arms, rewards)):
            arm = arm_reward_tuple[0]
            reward = arm_reward_tuple[1]
            self.rank_bandits[i].update(arm, reward, context=context)


class IndependentBandits(RankBandit):

    @staticmethod
    def find_next_possible_arm(selected_arms, ranked_arms):
        for arm in ranked_arms:
            if arm not in selected_arms:
                return arm

    def choose(self, context=None):
        selected_arms = []

        for i in range(self.num_ranks):
            selected_arm, ranked_arms = self.rank_bandits[i].choose()
            if selected_arm in selected_arms:
                selected_arm = self.find_next_possible_arm(selected_arms, ranked_arms)

            selected_arms.append(selected_arm)

        return selected_arms



from setuptools import find_packages, setup

with open('README.md') as f:
    long_description = f.read()

setup(
    name='mab-ranking',
    version='0.0.1',
    setup_cfg=True,
    python_requires='~=3.5',
    packages=find_packages(where='.'),
    long_description=long_description,
    long_description_content_type='text/markdown',
    setup_requires=['setuptools>=39.1.0'],
    url='https://github.com/Kenza-AI/mab-ranking',
    install_requires=[
        'boto3',
        'paramiko>=2.4.2, <2.4.99',
        'pathlib2>=2.3.0, <2.3.99',
        'requests>=2.20.0, <2.20.99',
        'sagemaker>=1.50.0, <1.50.99',
        'six>=1.10, <1.11.99',
        'future>=0.16.0, <0.17.99'
    ],
    test_suite='tests',
    zip_safe=True
)
